{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, header %}

{% block content %}
{% include "security/_messages.html" %}

{{ header(_("Login")) }}
<form action="{{ url_for_security('login') }}" method="POST" name="login_user_form">
  {{ login_user_form.hidden_tag() }}
  {{ render_field_with_errors(login_user_form.email) }}
  {{ render_field_with_errors(login_user_form.password) }}
  {{ render_field_with_errors(login_user_form.remember, field_class="checkbox") }}
  {{ render_field(login_user_form.submit, class_="button is-link") }}
</form>

{% endblock %}
